from abc import ABC, abstractmethod
from tinydb import TinyDB


class Strategy(ABC):

	def __init__(self, db_name: str, strategy_bot):
		'''
		Initializer for the Strategy class.

		:param str db_name: Required. The name of the database file to use, without the extension.
		:param autofold strategy_bot: Required. An instance of the autofold class.

		.. note::

			All child classes of Strategy are provided a local tinydb for non-volatile storage if needed.
			Note that tinydb is NOT threadsafe; access to a db should only occur from within the strategy that created it.
			Feel free to use your own storage medium as you see fit.

		Attributes:
		-----------
		- ``strategy_bot``: The ManifoldBot instance.
		- ``manifold_api``: The ManifoldAPI instance extracted from strategy_bot.
		- ``manifold_db_reader``: The ManifoldDatabaseReader instance extracted from strategy_bot.
		- ``manifold_subscriber``: The ManifoldSubscriber instance extracted from strategy_bot.
		- ``db``: The TinyDB instance for this strategy.
		''' 
		self.strategy_bot = strategy_bot
		self.manifold_api = strategy_bot.manifold_api
		self.manifold_db_reader = strategy_bot.manifold_db_reader
		self.manifold_subscriber = strategy_bot.manifold_subscriber
		self.db = TinyDB('dbs/'+db_name+'.json')
	
	@abstractmethod
	def start(self, *args, **kwargs):
		'''
		Abstract method to start the strategy.

		.. note::
			This method must be implemented in subclasses.

		:param args: Additional positional arguments.
		:param kwargs: Additional keyword arguments.
		'''
		pass
	
	@abstractmethod
	def stop(self, *args, **kwargs):
		'''
		Abstract method to stop the strategy.

		.. note::
			This method must be implemented in subclasses.

		:param args: Additional positional arguments.
		:param kwargs: Additional keyword arguments.
		'''
		pass



import json
import os
import time
from loguru import logger
import requests
from datetime import datetime
import threading
from queue import Queue, Full
from concurrent.futures import ThreadPoolExecutor, Future


DEV_DOMAIN = 'https://dev.manifold.markets' 
MAIN_DOMAIN = 'https://manifold.markets'

# Load API key from the environment variable
api_key = os.environ.get("MANIFOLD_API_KEY")
if api_key is None:
    logger.error("Environment variable MANIFOLD_API_KEY is not set")
	


'''
From docs:
Keep your bets to less than 10 per minute, amortized (transient spikes of over 10/min are okay).
Keep your reads to less than 100 per second.

A non-refundable transaction fee of M0.25 will be levied on any bet, sell, or limit order placed through the API, or by any account marked as a bot.
Comments placed through the API will incur a M1 transaction fee.
'''

'''
API interface is most recent with 2023-05-15 on changelog.
'''

READS_PER_SECOND = 100
BETS_PER_MINUTE = 5

class TokenBucket:
	def __init__(self, tokens, fill_rate):
		""" Tokens is the total tokens in the bucket. fill_rate is the rate in tokens/second."""
		self.capacity = tokens
		self._tokens = tokens
		self.fill_rate = fill_rate
		self.timestamp = time.time()

	def consume(self, tokens):
		""" Consume tokens from the bucket. Returns 0 if there are sufficient tokens, otherwise the expected time to wait in seconds."""
		if tokens > self._tokens:
			return (tokens - self._tokens) / self.fill_rate
		self._tokens -= tokens
		return 0

	def refill(self):
		""" Add new tokens to the bucket."""
		now = time.time()
		delta = self.fill_rate * (now - self.timestamp)
		self._tokens = min(self.capacity, self._tokens + delta)
		self.timestamp = now

class ManifoldAPI():
	def __init__(self, dev_mode=False):
		'''
        Initialize ManifoldAPI.

        :param bool dev_mode: Optional. Whether to enable developer mode. Default is False.
        
        .. note::
			- API key must be provided as an environment variable as MANIFOLD_API_KEY
            - Token buckets for reads and bets are initialized.
            - Request queues for reads and bets are set up.
            - Thread pool executor is set up.
            - Separate threads for processing read and bet queues are started.
        ''' 
		logger.debug("Initializing ManifoldAPI")
		self.dev_mode = dev_mode
		self._reads_bucket = TokenBucket(100, READS_PER_SECOND) 
		self._bets_bucket = TokenBucket(10, BETS_PER_MINUTE/60.0) 
		self._bets_queue = Queue(maxsize=1000)
		self._reads_queue = Queue(maxsize=1000) 
  
		self._executor = ThreadPoolExecutor(thread_name_prefix="MF_API", max_workers=5)
	
		self.running = True
  
		# Start processing the request queues in a separate thread
		self._read_thread = threading.Thread(target=self._process_read_queue, daemon=True)
		self._read_thread.start()

		# Start processing the bets queues in a separate thread
		self._bet_thread = threading.Thread(target=self._process_bet_queue, daemon=True)
		self._bet_thread.start()

	def shutdown(self):
		'''
        Shutdown the ManifoldAPI.

        .. note::
            - Stops all running threads.
            - Shuts down the thread pool _executor.
            - Sets all pending Futures to exceptions stating "API is shutting down".
            
        :return: None
        :rtype: None
        ''' 
		logger.info("Shutting down ManifoldAPI") 
		self.running = False
		self._read_thread.join()
		self._bet_thread.join()
		self._executor.shutdown(wait=False)
  
		# Set futures to exceptions
		while not self._reads_queue.empty():
			_, _, _, future = self._reads_queue.get()
			future.set_exception(Exception("API is shutting down"))
		while not self._bets_queue.empty():
			_, _, _, future = self._bets_queue.get()
			future.set_exception(Exception("API is shutting down")) 

	def _make_request(self, endpoint, method="GET", params=None, future=None):
		headers = {"Authorization": f"Key {api_key}"}
		log_data = {
				"timestamp": datetime.now().isoformat(),
				"endpoint": endpoint,
				"method": method,
				"params": params
			}
		logger.debug(f"API call: {json.dumps(log_data)}")

		if not self.dev_mode:
			endpoint = MAIN_DOMAIN + endpoint
		else:
			endpoint = DEV_DOMAIN + endpoint

		try:
			if method == "GET":
				response = requests.get(endpoint, headers=headers, params=params, timeout=5)
			elif method == "POST":
				response = requests.post(endpoint, headers=headers, json=params, timeout=5)

			if response.status_code != 200:
				logger.error(f"Error in API call: {response.status_code}, {response.json}")
				if future:
					future.set_exception(Exception(f"HTTP Error: {response.status_code}"))
			else:
				if future:
					future.set_result(response.json())
		except Exception as e:
			logger.error(f"Exception occurred while making a request: {e}") 
			if future:
				future.set_exception(e)
				
	def _process_read_queue(self):
		while self.running:

			self._reads_bucket.refill() 

			# Process read requests
			while not self._reads_queue.empty():
				wait_time = self._reads_bucket.consume(1)
				if wait_time > 0:
					time.sleep(wait_time)
				endpoint, method, params, future = self._reads_queue.get()
				self._executor.submit(self._make_request, endpoint, method, params, future)

			time.sleep(0.1)  # Sleep 100 milliseconds if the queue is empty 


	def _process_bet_queue(self):
		while self.running:
      
			self._bets_bucket.refill() 

			# Process bet requests if we can
			while not self._bets_queue.empty():
				wait_time = self._bets_bucket.consume(1)
				if wait_time > 0:
					time.sleep(wait_time)
				endpoint, method, params, future = self._bets_queue.get()
				self._executor.submit(self._make_request, endpoint, method, params, future)
    
			time.sleep(0.1)  # Sleep 100 milliseconds if the queue is empty 
    

	def retrieve_all_data(self, api_call_func, max_limit=1000, **api_params):
		'''
		Iteratively retrieves all available data from an API endpoint that supports pagination via a `before` parameter.

		:param Callable api_call_func: 
			Required. A function that makes the API call and returns a Future object.
		:param int max_limit: 
			Optional. The maximum number of items to request in a single API call. Default is 1000.
		:param api_params: 
			Optional. Additional parameters to pass to the API call function. Must be passed as keyword arguments.
		:type api_params: dict
		:return: 
			A list containing all retrieved items.
		:rtype: list

		.. note::

			This function is blocking.

		**Example**

		.. code-block:: python

			retrieve_all_data(api_function, max_limit=200, param1="value1", param2="value2")

		'''
  
		logger.debug(f"Starting retrieve_all_data") 
		all_data = []
		last_item_id = None
		has_more_data = True
		
		while has_more_data:
			if not self.running:
				raise Exception("retrieve_all_data interrupted by thread exit")
			try:
				# Include the 'before' parameter if we have a last_item_id to work with
				if last_item_id:
					api_params['before'] = last_item_id

				# Make the API call
				future_response = api_call_func(limit=max_limit, **api_params)
				response = future_response.result()

				if response:
					all_data.extend(response)
					last_item_id = response[-1]['id']
					has_more_data = len(response) == max_limit
				else:
					has_more_data = False
			except Exception as e:
				logger.error(f"An error occurred during retrieve_all_data: {e}")
				has_more_data = False

		return all_data   

	def get_user_by_username(self, username):
		'''
		GET /v0/user/[username]

		Gets a user by their username. Remember that usernames may change.

		:param str username: Required. The username of the user.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
	 
		future = Future()
	
		self._reads_queue.put((f"/api/v0/user/{username}", "GET", None, future))
		return future

	def get_user_by_id(self, userId):
		'''
		GET /v0/user/by-id/[id]

		Gets a user by their unique ID. Many other API endpoints return this as the userId.

		:param str userId: Required. The ID of the user.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
	 
		future = Future()
		
		self._reads_queue.put((f"/api/v0/user/by-id/{userId}", "GET", None, future))
		return future

	def get_me(self):
		'''
		GET /v0/me

		Returns the authenticated user.

		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''

		future = Future()
  
		self._reads_queue.put((f"/api/v0/me", "GET", None, future))
		return future

	# Returns 500 error	
	# def get_groups(self):
	# 	'''
	# 	GET /v0/groups

	# 	Gets all groups, in no particular order.

	# 	Parameters:
	# 	availableToUserId: Optional. if specified, only groups that the user can join and groups they've already joined will be returned.
	# 	Requires no authorization.
	# 	'''
	 
	# 	future = Future()
		
	# 	self._reads_queue.put(("/api/v0/groups", "GET", None, future))
	# 	return future

	def get_group_by_slug(self, groupSlug):
		'''
		GET /v0/group/[slug]

		Gets a group by its slug.
		
		Note: group is singular in the URL.

		:param str groupSlug: Required. The slug of the group.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
	 
		future = Future()
		
		self._reads_queue.put((f"/api/v0/group/{groupSlug}", "GET", None, future))
		return future

	def get_group_by_id(self, groupId):
		'''
		GET /v0/group/by-id/[id]

		Gets a group by its unique ID.

		:param str groupId: Required. The id of the group.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
	 
		future = Future()
		
		self._reads_queue.put((f"/api/v0/group/by-id/{groupId}", "GET", None, future))
		return future
 
	def get_group_markets_by_id(self, groupId):
		'''
		GET /v0/group/by-id/[id]/markets

		Gets a group's markets by its unique ID.

		:param str groupId: Required. The id of the group.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
		future = Future()
		
		self._reads_queue.put((f"/api/v0/group/by-id/{groupId}/markets", "GET", None, future))
		return future

	def get_markets(self, limit=500, before=None):
	 
		'''
		GET /v0/markets

		Lists all markets, ordered by creation date descending.

		:param int limit: Optional. How many markets to return. The maximum is 1000 and the default is 500.
		:param str before: Optional. The ID of the market before which the list will start.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
		future = Future()
		
		params = {"limit": limit}
		if before:
			params["before"] = before
		self._reads_queue.put((f"/api/v0/markets", "GET", params, future))
		return future
	
	def get_market_by_id(self, marketId):
		'''
		GET /v0/market/[marketId]

		Gets information about a single market by ID. Includes answers, but not bets and comments. Use /bets or /comments with a market ID to retrieve bets or comments.

		:param str marketId: Required. The ID of the market.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
	 
		future = Future()
		
		self._reads_queue.put((f"/api/v0/market/{marketId}", "GET", None, future))
		return future

	def get_market_positions(self, marketId, order='profit', top=None, bottom=None, userId=None):
		'''
		NOTE: This API endpoint will break for markets with > 4650 positions. Setting either the top or bottom parameters is required to mitigate a 500 server error.
		See https://github.com/manifoldmarkets/manifold/issues/2031 

		GET /v0/market/[marketId]/positions

		Get positions information about a single market.

		:param str marketId: Required. The ID of the market.
		:param str order: Optional. The field to order results by. Default is 'profit'. Options are 'shares' or 'profit'.
		:param int top: Optional. The number of top positions (ordered by 'order') to return. Default is None.
		:param int bottom: Optional. The number of bottom positions (ordered by 'order') to return. Default is None.
		:param str userId: Optional. The user ID to query by. Default is None. If provided, only the positions for this user will be returned.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
		future = Future()
		
		params = {"order": order}
		if top:
			params["top"] = top
		if bottom:
			params["bottom"] = bottom
		if userId:
			params["userId"] = userId
   
		self._reads_queue.put((f"/api/v0/market/{marketId}/positions", "GET", params, future))
		return future

	def get_market_by_slug(self, marketSlug):
		'''
		GET /v0/slug/[marketSlug]
	
		Gets information about a single market by slug (the portion of the URL path after the username).

		:param str marketSlug: Required. The slug of the market.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
		future = Future()
		self._reads_queue.put((f"/api/v0/slug/{marketSlug}", "GET", None, future))
		return future


	def search_markets(self, terms=None):
		'''
		GET /v0/search-markets

		Search markets by keywords, limited to 100 results.

		:param str terms: Optional. A space-separated list of keywords to search for.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''

		future = Future()
		params = {}
		if terms:
			params["terms"] = terms
   
		self._reads_queue.put((f"/api/v0/search-markets", "GET", params, future))
		return future	

	def get_users(self, limit=None, before=None):
		'''
		GET /v0/users

		Lists all users, ordered by creation date descending.

		:param int limit: Optional. How many users to return. The maximum is 1000 and the default is 500.
		:param str before: Optional. The ID of the user before which the list will start. For example, if you ask for the most recent 10 users, and then perform a second query for 10 more users with before=[the id of the 10th user], you will get users 11 through 20.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future
		'''

		future = Future()
		params = {}
		if limit:
			params["limit"] = limit 
		if before:
			params["before"] = before
   
		self._reads_queue.put((f"/api/v0/users", "GET", params, future))
		return future	

	def make_bet(self, amount, contractId, outcome, limitProb=None, expiresAt=None):
		'''
		POST /v0/bet

		Places a new bet on behalf of the authorized user.

		:param float amount:
			Required. The amount to bet, in mana, before fees.
		:param str contractId:
			Required. The ID of the contract (market) to bet on. 
		:param str outcome:
			Required. The outcome to bet on. The outcome type is market-specific:
			
			- For binary markets: "YES" or "NO"
			- For free-response markets: ID of the free response answer
			- For numeric markets: String representing the target bucket
			
		:param float limitProb:
			Optional. A number between 0.01 and 0.99 inclusive representing the limit probability for your bet. 
			
			- For example, if the current market probability is 50%:
				- A M$10 bet on "YES" with ``limitProb=0.4`` would not be filled until the market probability moves down to 40%.
				- A M$100 bet on "YES" with ``limitProb=0.6`` would fill partially or completely depending on current market conditions.
				
			Any remaining unfilled bet will remain as an open offer for future matches.
			
		:param int expiresAt:
			Optional. A Unix timestamp (in milliseconds) at which the limit bet should be automatically canceled.

		:return:
			A Future object representing the eventual result of the API call.
		:rtype: Future

		**Examples**

		.. code-block:: python

			make_bet(100, "contractId123", "YES", limitProb=0.6)
		'''

		future = Future()
		params = {"amount": amount, "contractId": contractId, "outcome": outcome}
		if limitProb:
			params["limitProb"] = limitProb
		if expiresAt:
			params["expiresAt"] = expiresAt
   
		self._bets_queue.put((f"/api/v0/bet", "POST", params, future))
		return future	

	def cancel_bet(self, betId):
		'''
		POST /v0/bet/cancel/[betId]

		Cancels the limit order of a bet with the specified id.

		:param str betId: 
			Required. The unique identifier for the bet to be cancelled.
			
		:return:
			A Future object representing the eventual result of the API call.
		:rtype: Future

		.. note:: 
			This action is irreversible.

		**Examples**

		.. code-block:: python

			cancel_bet("betId123")
		'''
		future = Future()
		self._bets_queue.put((f"/api/v0/bet/cancel/{betId}", "POST", None, future))
		return future

	def create_market(self, outcomeType, question, description=None, closeTime=None, visibility=None, groupId=None, initialProb=None, min=None, max=None, isLogScale=None, initialValue=None, answers=None):
		'''
		POST /v0/market

		Creates a new market on behalf of the authorized user.

		:param str outcomeType: Required. The type of outcome for the market.
		:param str question: Required. The main question for the market.
		:param str description: Optional. A detailed description for the market.
		:param int closeTime: Optional. Time when the market closes.
		:param str visibility: Optional. The visibility setting for the market.
		:param str groupId: Optional. The group ID associated with the market.
		:param float initialProb: Optional. The initial probability for the market outcome.
		:param float min: Optional. The minimum value for a numeric market.
		:param float max: Optional. The maximum value for a numeric market.
		:param bool isLogScale: Optional. Whether the market uses a logarithmic scale.
		:param float initialValue: Optional. The initial value for the market.
		:param list[str] answers: Optional. Possible answers for a free-response market.
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future

		**Examples**

		.. code-block:: python

		   create_market("binary", "Will it rain?", description="Weather prediction", closeTime=1633027200, visibility="public")
		'''
		future = Future()
		params = {"outcomeType": outcomeType, "question": question}
		if description: params["description"] = description
		if closeTime: params["closeTime"] = closeTime
		if visibility: params["visibility"] = visibility
		if groupId: params["groupId"] = groupId
		if initialProb: params["initialProb"] = initialProb
		if min: params["min"] = min
		if max: params["max"] = max
		if isLogScale: params["isLogScale"] = isLogScale
		if initialValue: params["initialValue"] = initialValue
		if answers: params["answers"] = answers

		self._bets_queue.put((f"/api/v0/market", "POST", params, future))
		return future

	def add_liquidity(self, marketId, amount):
		'''
		POST /v0/market/[marketId]/add-liquidity

		Adds liquidity to a specific market.

		:param str marketId: 
			The ID of the market. This parameter is required.
		:param float amount: 
			The amount of liquidity to be added. This parameter is required.

		:return: 
			A Future object representing the eventual result of the API call.
		:rtype: Future

		**Examples**

		.. code-block:: python

			add_liquidity("marketId123", 500.0)
		'''
		future = Future()
		params = {"amount": amount}
		self._bets_queue.put((f"/api/v0/market/{marketId}/add-liquidity", "POST", params, future))
		return future

	def close_market(self, marketId, closeTime=None):
		'''
		POST /v0/market/[marketId]/close
		
		Closes a market on behalf of the authorized user.
		
		:param str marketId: 
			The unique identifier for the market to be closed. This parameter is required.
		:param int closeTime: 
			Optional. Milliseconds since the epoch to close the market at. If not provided, the market will be closed immediately. Cannot provide close time in the past.
		:return: 
			A Future object representing the eventual result of the API call.
		:rtype: Future
		
		**Examples**

		.. code-block:: python

			close_market("marketId123")
			close_market("marketId123", closeTime=1672444800000)
		'''
		future = Future()
		params = {}
		if closeTime: params["closeTime"] = closeTime
		self._bets_queue.put((f"/api/v0/market/{marketId}/close", "POST", params, future))
		return future

	def manage_group_market(self, marketId, groupId, remove=None):
		'''
		POST /v0/market/[marketId]/group

		Add or remove a market to/from a group.

		:param str marketId: Required. The ID of the market.
		:param str groupId: Required. The ID of the group. Must be an admin, moderator, or creator of the group if curated/private. Must be the market creator or trustworthy-ish if the group is public.
		:param bool remove: Optional. Set to true to remove the market from the group.
		
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future

		**Example Usage**

		.. code-block:: python

			manage_group_market("marketId123", "groupId456", remove=True)
		'''
		future = Future()
		params = {"groupId": groupId}
		if remove: params["remove"] = remove
		self._bets_queue.put((f"/api/v0/market/{marketId}/group", "POST", params, future))
		return future

	def resolve_market(self, marketId, outcome, probabilityInt=None, resolutions=None, value=None):
		'''
		POST /v0/market/[marketId]/resolve

		Resolves a market on behalf of the authorized user.

		:param str marketId: 
			Required. The ID of the market.

		:param str outcome: 
			Required. Outcome varies based on the type of market.
			
			- For binary markets: One of "YES", "NO", "MKT", or "CANCEL".
			- For free response or multiple choice markets: One of "MKT", "CANCEL", or a number indicating the answer index.
			- For numeric markets: One of "CANCEL", or a number indicating the selected numeric bucket ID.

		:param int probabilityInt: 
			Optional. The probability to use for MKT resolution in binary markets.
			Also, required if `value` is present in numeric markets.

		:param list[dict] resolutions:
			Optional. An array of {answer, pct} objects to use as the weights for resolving in favor of multiple free response options.
			Can only be set with "MKT" outcome. Note that the total weights must add to 100.

		:param int value: 
			Optional. The value that the market resolves to in numeric markets.

		:raises ValueError: 
			If the total weights in `resolutions` do not add up to 100.
			
		**Examples**

		.. code-block:: python

			resolve_market("marketId123", "YES", probabilityInt=80)
			resolve_market("marketId456", "MKT", resolutions=[{"answer": "A", "pct": 40}, {"answer": "B", "pct": 60}])
		'''
		future = Future()
		params = {"outcome": outcome}
		if probabilityInt: params["probabilityInt"] = probabilityInt
		if resolutions: params["resolutions"] = resolutions
		if value: params["value"] = value

		self._bets_queue.put((f"/api/v0/market/{marketId}/resolve", "POST", params, future))
		return future

	def sell_shares(self, marketId, outcome=None, shares=None):
		'''
		POST /v0/market/[marketId]/sell

		Sells some quantity of shares in a binary market.

		:param str marketId: The unique identifier for the binary market where shares are being sold. This parameter is required.
		:param str outcome: Optional. Specifies the outcome for which shares are being sold. Can be "YES" or "NO".
		:param float shares: Optional. The amount of shares to sell for the given outcome. If not provided, all shares owned will be sold.
		
		:return: A Future object representing the eventual result of the API call.
		:rtype: Future

		**Examples**

		.. code-block:: python

			sell_shares("marketId123", "YES", 10)
		'''
		future = Future()
		params = {}
		if outcome:
			params["outcome"] = outcome
		if shares:
			params["shares"] = shares
		self._bets_queue.put((f"/api/v0/market/{marketId}/sell", "POST", params, future))
		return future

	def create_comment(self, contractId, content=None, html=None, markdown=None):
		'''
		POST /v0/comment
		
		Creates a comment in the specified market.

		:param str contractId:
			Required. The ID of the market to comment on.
		:param content:
			Optional. The comment to post, formatted as TipTap json.
			:conflicted: html, markdown
		:type content: str or None
		:param html:
			Optional. The comment to post, formatted as an HTML string.
			:conflicted: content, markdown
		:type html: str or None
		:param markdown:
			Optional. The comment to post, formatted as a markdown string.
			:conflicted: content, html
		:type markdown: str or None
		:return:
			A Future object representing the eventual result of the API call.
		:rtype: Future
		
		.. note::
		   You should provide either `content`, `html`, or `markdown` but not multiple at the same time. They are mutually exclusive.

		**Examples**

		.. code-block:: python

			create_comment("contractId123", content="some TipTap json content")
			create_comment("contractId123", html="<p>some HTML content</p>")
			create_comment("contractId123", markdown="## some markdown content")
		'''
		future = Future()
		params = {"contractId": contractId}
		if content:
			params["content"] = content
		elif html:
			params["html"] = html
		elif markdown:
			params["markdown"] = markdown
		self._reads_queue.put((f"/api/v0/comment", "POST", params, future))
		return future

	def get_comments(self, contractId=None, contractSlug=None):
		'''
		GET /v0/comments
		
		Gets a list of comments for a contract.

		:param str contractId: Optional. The ID of the contract to read comments for.
			Either an ID or a slug must be specified.
		:param str contractSlug: Optional. The slug of the contract to read comments for.
			Either an ID or a slug must be specified.

		:return: A Future object representing the eventual result of the API call.
		:rtype: Future

		.. note::
		   Either `contractId` or `contractSlug` must be specified.

		**Examples**

		.. code-block:: python

		   get_comments(contractId="someContractId")
		   get_comments(contractSlug="someContractSlug")
		'''
		future = Future()
		params = {}
		if contractId:
			params["contractId"] = contractId
		if contractSlug:
			params["contractSlug"] = contractSlug
		self._reads_queue.put((f"/api/v0/comments", "GET", params, future))
		return future

	def get_bets(self, userId=None, username=None, contractId=None, contractSlug=None, limit=None, before=None):
		'''
		GET /v0/bets

		Retrieves a list of bets, sorted by their creation date in descending order.

		:param str userId: 
			Optional. If provided, returns only bets created by the user with this ID.
		:param str username: 
			Optional. If provided, returns only bets created by the user with this username.
		:param str contractId: 
			Optional. If provided, returns only bets associated with this contract ID.
		:param str contractSlug: 
			Optional. If provided, returns only bets associated with this contract slug.
		:param int limit: 
			Optional. The number of bets to return. Defaults to and maxes out at 1000.
		:param str before: 
			Optional. Specifies the ID of the bet to start the list from, effectively functioning as an offset. 
			For instance, after requesting the 10 most recent bets, supplying the ID of the 10th bet in a new query would yield bets 11 through 20.
			
		:return: 
			A Future object representing the eventual result of the API call.
		:rtype: Future
		'''
		future = Future()
		params = {}
		if userId:
			params["userId"] = userId
		if username:
			params["username"] = username
		if contractId:
			params["contractId"] = contractId
		if contractSlug:
			params["contractSlug"] = contractSlug
		if limit:
			params["limit"] = limit
		if before:
			params["before"] = before
		self._reads_queue.put((f"/api/v0/bets", "GET", params, future))
		return future




from loguru import logger
import threading
import queue
import concurrent.futures
import sqlite3
import threading
import time

from autofold.utils.str_utils import collapse_list_of_strings_to_string
import concurrent.futures

# Helper function for multiple upserts
def sanitize_value(value):
        if isinstance(value, float):
            return round(value, 4) 
        elif isinstance(value, int) and (value > 2**63 - 1 or value < -2**63):
            logger.error(f"Integer {value} is too large to store in SQLite. Setting it to 2**63 - 1")
            return 2**63 - 1
        # You can add more sanitation logic as needed
        return value

def prepare_and_execute_multi_upsert(conn, query, fields, data):
    query_fields = ", ".join(fields)
    query_placeholders = ", ".join("?" for _ in fields)
    sql_query = query.format(fields=query_fields, placeholders=query_placeholders)
    values_tuple = [tuple(sanitize_value(data.get(field, None)) for field in fields) for data in data]
    conn.executemany(sql_query, values_tuple)

# Helper function for multiple deletions
def prepare_and_execute_multi_deletion(conn, query, ids):
    # Check if the first item in ids is a tuple
    if isinstance(ids[0], tuple):
        values_tuple = ids
    else:
        values_tuple = [(item,) for item in ids]

    conn.executemany(query, values_tuple)
    



class ManifoldDatabase:
    def __init__(self):
        self.local_storage = threading.local()

    def get_conn(self):
        if not hasattr(self.local_storage, "conn"):
            self.local_storage.conn = sqlite3.connect("dbs/manifold_database.db")
            self.local_storage.conn.execute("PRAGMA journal_mode=WAL;")
        return self.local_storage.conn

    def create_tables(self):
        conn = self.get_conn()
        logger.debug("Creating tables") 
        '''
        ########################################################
        ####                    USERS                       ####
        ########################################################
        '''
        # Create users table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS users (
            id TEXT PRIMARY KEY,
            createdTime INTEGER,
            name TEXT,
            username TEXT,
            url TEXT,
            bio TEXT,
            balance REAL,
            totalDeposits REAL,
            totalPnLCached REAL,
            retrievedTimestamp INTEGER
        );
        """)
        
        '''
        ########################################################
        ####                    MARKETS                     ####
        ########################################################
        '''
        # Create binary markets table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS binary_choice_markets (
            id TEXT PRIMARY KEY,
            closeTime INTEGER,
            createdTime INTEGER,
            creatorId TEXT,
            creatorName TEXT,
            creatorUsername TEXT,
            isResolved BOOLEAN,
            lastUpdatedTime INTEGER,
            mechanism TEXT,
            outcomeType TEXT,
            p REAL,
            probability REAL,
            question TEXT,
            textDescription TEXT,
            totalLiquidity REAL,
            url TEXT,
            volume REAL,
            volume24Hours REAL,
            pool_NO REAL,
            pool_YES REAL,
            groupSlugs TEXT,
            retrievedTimestamp INTEGER,
            lite INTEGER
        );
        """)
        
        # Create multiple choice markets table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS multiple_choice_markets (
            id TEXT PRIMARY KEY,
            closeTime INTEGER,
            createdTime INTEGER,
            creatorId TEXT,
            creatorName TEXT,
            creatorUsername TEXT,
            isResolved BOOLEAN,
            lastUpdatedTime INTEGER,
            mechanism TEXT,
            outcomeType TEXT,
            question TEXT,
            textDescription TEXT,
            totalLiquidity REAL,
            volume REAL,
            volume24Hours REAL,
            url TEXT,
            groupSlugs TEXT,
            retrievedTimestamp INTEGER,
            lite INTEGER
        );
        """)

        # Create 'nested' answers table for multiple choice markets
        conn.execute("""
        CREATE TABLE IF NOT EXISTS multiple_choice_market_answers (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            contractId TEXT,
            createdTime INTEGER,
            fsUpdatedTime TEXT,
            isOther INTEGER,
            answerIndex INTEGER,
            probability REAL,
            subsidyPool REAL,
            text TEXT,
            totalLiquidity REAL,
            userId TEXT,
            pool_NO REAL,
            pool_YES REAL,
            FOREIGN KEY(contractId) REFERENCES multiple_choice_markets(id)
        ); 
        """)

        '''
        ########################################################
        ####                 CONTRACT METRICS               ####
        ########################################################
        '''
        # Create contract_metrics table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS contract_metrics (
            contractId TEXT,
            hasNoShares INTEGER,
            hasShares INTEGER,
            hasYesShares INTEGER,
            invested REAL,
            loan REAL,
            maxSharesOutcome TEXT,
            payout REAL,
            profit REAL,
            profitPercent REAL,
            userId TEXT,
            userUsername TEXT,
            userName TEXT,
            lastBetTime INTEGER,
            retrievedTimestamp INTEGER,
            PRIMARY KEY (contractId, userId) 
        );
        """)
        
        conn.execute("""
        CREATE TABLE IF NOT EXISTS contract_metrics_from (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            contractId TEXT,
            userId TEXT,
            period TEXT,
            value REAL,
            profit REAL,
            invested REAL,
            prevValue REAL,
            profitPercent REAL,
            FOREIGN KEY (contractId, userId) REFERENCES contract_metrics(contractId, userId)
        ); 
        """)

        # contract_metrics_totalShares table to represent 'totalShares' nested structure
        conn.execute("""
        CREATE TABLE IF NOT EXISTS contract_metrics_totalShares (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            contractId TEXT,
            userId TEXT,
            outcome TEXT,
            numberOfShares REAL,
            FOREIGN KEY (contractId, userId) REFERENCES contract_metrics(contractId, userId)
        ); 
        """)

        '''
        ########################################################
        ####                    BETS                        ####
        ########################################################
        '''
        # Create bets table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS bets (
            id TEXT PRIMARY KEY,
            userId TEXT,
            contractId TEXT,
            isFilled INTEGER,
            amount REAL,
            probBefore REAL,
            isCancelled INTEGER,
            outcome TEXT,
            shares REAL,
            limitProb REAL,
            loanAmount REAL,
            orderAmount REAL,
            probAfter REAL,
            createdTime INTEGER,
            retrievedTimestamp INTEGER
        );
        """)

        # Create fees table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS bet_fees (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            betId TEXT,
            creatorFee REAL,
            liquidityFee REAL,
            platformFee REAL,
            FOREIGN KEY (betId) REFERENCES bets(id)
        );
        """)

        # Create fills table
        conn.execute("""
        CREATE TABLE IF NOT EXISTS bet_fills (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            betId TEXT,
            timestamp INTEGER,
            matchedBetId TEXT,
            amount REAL,
            shares REAL,
            FOREIGN KEY (betId) REFERENCES bets(id)
        );
        """)

        conn.commit()
        
    '''
    ########################################################
    ####                    USERS                       ####
    ########################################################
    '''

    def upsert_users(self, users: list[dict]):
        # Get database connection
        conn = self.get_conn()
        
        logger.debug(f"Upserting {len(users)} users")
        
        # Current UNIX epoch time for all markets in this batch
        current_time = int(time.time())
        
        # Begin transaction
        conn.execute("BEGIN TRANSACTION;")
        
        try:
            # Base table fields
            base_fields = [
                    "id", "createdTime", "name", "username", 
                    "url", "bio", "balance", "totalDeposits", 
                    "totalPnLCached", "retrievedTimestamp"
                ]
            
            # Insert or Replace into the base table
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO users ({fields}) VALUES ({placeholders})",
                fields=base_fields,
                data=[
                    {**user, 
                     "retrievedTimestamp": current_time,
                    } for user in users],
            )
            
            
            # Commit transaction
            conn.commit()
    
        except sqlite3.Error as e:
            logger.error("Database error in upset_users:", e)
            conn.rollback()  # Rollback transaction in case of error        

        logger.debug("Upsert users successful")


    
    '''
    ########################################################
    ####                    MARKETS                     ####
    ########################################################
    '''

    # Upsert Market
    def upsert_binary_choice_markets(self, markets: list[dict], lite=True):
        # Get database connection
        conn = self.get_conn()
        
        logger.debug(f"Upserting {len(markets)} binary choice markets")
        
        # Current UNIX epoch time for all markets in this batch
        current_time = int(time.time())
        
        # Begin transaction
        conn.execute("BEGIN TRANSACTION;")
        
        try:
            # Base table fields
            base_fields = [
                "id", "closeTime", "createdTime", "creatorId", "creatorName", 
                "creatorUsername", "isResolved", "lastUpdatedTime", "mechanism", 
                "outcomeType", "p", "probability", "question", "textDescription", 
                "totalLiquidity", "volume", "volume24Hours", "url", "pool_NO",
                "pool_YES", "groupSlugs", "retrievedTimestamp", "lite"
            ]
            
            # Insert or Replace into the base table
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO binary_choice_markets ({fields}) VALUES ({placeholders})",
                fields=base_fields,
                data=[
                    {**market, 
                    "retrievedTimestamp": current_time,
                    "lite": int(lite),
                    "groupSlugs": collapse_list_of_strings_to_string(market.get("groupSlugs", "")),
                    "pool_NO": market.get("pool", {}).get("NO", None),
                    "pool_YES": market.get("pool", {}).get("YES", None)
                    } for market in markets],
            )
            
            # Commit transaction
            conn.commit()
    
        except sqlite3.Error as e:
            logger.error("Database error in upsert_binary_choice_markets:", e)
            conn.rollback()  # Rollback transaction in case of error

        logger.debug("Upsert binary choice markets successful")


    def upsert_multiple_choice_markets(self, markets: list[dict], lite=True):
        # Get database connection
        conn = self.get_conn()
        
        logger.debug(f"Upserting {len(markets)} markets")
        
        # Current UNIX epoch time for all markets in this batch
        current_time = int(time.time())
        
        # Begin transaction
        conn.execute("BEGIN TRANSACTION;")
        
        try:
            # Base table fields
            base_fields = [
                "id", "closeTime", "createdTime", "creatorId", "creatorName", 
                "creatorUsername", "isResolved", "lastUpdatedTime", "mechanism", 
                "outcomeType", "question", "textDescription", 
                "totalLiquidity", "volume", "volume24Hours", 
                "url", "groupSlugs", "retrievedTimestamp", "lite"
            ]
            
            # Insert or Replace into the base table
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO multiple_choice_markets ({fields}) VALUES ({placeholders})",
                fields=base_fields,
                data=[
                    {**market, 
                     "retrievedTimestamp": current_time,
                     "lite": int(lite),
                     "groupSlugs": collapse_list_of_strings_to_string(market.get("groupSlugs", ""))
                    } for market in markets],
            )
            
            # Handle nested tables (answers)
            if not lite:
                
                # Delete existing answers for the markets
                prepare_and_execute_multi_deletion(
                    conn=conn,
                    query="DELETE FROM multiple_choice_market_answers WHERE contractId = ?",
                    ids=[market["id"] for market in markets]
                )
                
                answer_fields = [
                    "contractId", "createdTime", "fsUpdatedTime", "isOther", "answerIndex", 
                    "probability", "subsidyPool", "text", "totalLiquidity", "userId", 
                    "pool_NO", "pool_YES"
                ]
                
                prepare_and_execute_multi_upsert(
                    conn=conn,
                    query="INSERT OR REPLACE INTO multiple_choice_market_answers ({fields}) VALUES ({placeholders})",
                    fields=answer_fields,
                    data=[
                        {**answer,
                         "pool_NO": answer.get("pool", {}).get("NO", None),
                         "pool_YES": answer.get("pool", {}).get("YES", None)
                         } for market in markets for answer in market.get("answers", [])]
                )
            
            # Commit transaction
            conn.commit()
            
    
        except sqlite3.Error as e:
            logger.error("Database error in upsert_multiple_choice_markets:", e)
            conn.rollback()  # Rollback transaction in case of error
            
        logger.debug("Upsert multiple choice markets successful")
        

        

    '''
    ########################################################
    ####                 CONTRACT METRICS               ####
    ########################################################
    '''
    def upsert_contract_metrics(self, contract_metrics: list[dict]):
        # Get database connection
        conn = self.get_conn()
        
        logger.debug(f"Upserting {len(contract_metrics)} contract metrics")
        
        # Current UNIX epoch time for all contract_metrics in this batch
        current_time = int(time.time())
        
        # Begin transaction for better performance and data integrity
        conn.execute("BEGIN TRANSACTION;")

        try:

            # Base table
            base_fields = [
                "contractId", "hasNoShares", "hasShares", "hasYesShares",
                "invested", "loan", "maxSharesOutcome", "payout", 
                "profit", "profitPercent", "userId", "userUsername", 
                "userName", "lastBetTime", "retrievedTimestamp"
            ]
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO contract_metrics ({fields}) VALUES ({placeholders})",
                fields=base_fields,
                data=[
                    {**metric, "retrievedTimestamp": current_time} for metric in contract_metrics]
            )

            
            # Handle nested tables (from and totalShares)
            # Delete entries
            prepare_and_execute_multi_deletion(
                conn=conn,
                query="DELETE FROM contract_metrics_from WHERE contractId = ? AND userId = ?",
                ids=[(contract_metric["contractId"], contract_metric["userId"]) for contract_metric in contract_metrics]
            )
            
            from_fields = ["contractId", "userId", "period", "value", "profit", "invested", "prevValue", "profitPercent"]
            
            # Clean data (some 'from' entries report profit percents of over a trillion, because of buggy invested value tracking. Set anything over 1,000,000% to -1)
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO contract_metrics_from ({fields}) VALUES ({placeholders})",
                fields=from_fields,
                data=[
                    {**from_vals,
                        "profitPercent": 1_000_000 if contract_metric.get("profitPercent", 0) > 1_000_000 else contract_metric.get("profitPercent", 0),
                        "contractId": contract_metric.get("contractId", None),
                        "userId": contract_metric.get("userId", None), 
                        "period": period} for contract_metric in contract_metrics for period, from_vals in contract_metric.get("from", {}).items()]
            )

            # Delete entries
            prepare_and_execute_multi_deletion(
                conn=conn,
                query="DELETE FROM contract_metrics_totalShares WHERE contractId = ? AND userId = ?",
                ids=[(contract_metric["contractId"], contract_metric["userId"]) for contract_metric in contract_metrics]
            )
            
            total_shares_fields = ["contractId", "userId", "outcome", "numberOfShares"]
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO contract_metrics_totalShares ({fields}) VALUES ({placeholders})",
                fields=total_shares_fields,
                data=[
                    {
                        "contractId": contract_metric.get("contractId", None),
                        "userId": contract_metric.get("userId", None),
                        "outcome": outcome,
                        "numberOfShares": numberOfShares
                        } for contract_metric in contract_metrics for outcome, numberOfShares in contract_metric.get("totalShares", {}).items()]
            )
        
            # Commit transaction
            conn.commit()
    
        except sqlite3.Error as e:
            logger.error("Database error in upsert_contract_metrics:", e)
            conn.rollback()  # Rollback transaction in case of error

        logger.debug("Upsert contract metrics successful")

    '''
    ########################################################
    ####                      BETS                      ####
    ########################################################
    '''
    def upsert_bets(self, bets: list[dict]):
        # Get database connection
        conn = self.get_conn()
        
        logger.debug(f"Upserting {len(bets)} bets")
        
        # Current UNIX epoch time for all bets in this batch
        current_time = int(time.time())
        
        # Begin transaction for performance and data integrity
        conn.execute("BEGIN TRANSACTION;")

        try:
            # Fields for the bets base table
            bet_fields = [
                "id", "userId", "contractId", "isFilled", "amount", "probBefore",
                "isCancelled", "outcome", "shares", "limitProb", "loanAmount", 
                "orderAmount", "probAfter", "createdTime", "retrievedTimestamp"
            ]
            
            # Insert or replace into the bets base table
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO bets ({fields}) VALUES ({placeholders})",
                fields=bet_fields,
                data=[{**bet, "retrievedTimestamp": current_time} for bet in bets]
            )
            
            # Handle nested tables (fees and fills)
            # Delete entries
            prepare_and_execute_multi_deletion(
                conn=conn,
                query="DELETE FROM bet_fees WHERE betId = ?",
                ids=[bet["id"] for bet in bets]
            )

            fee_fields = ["betId", "creatorFee", "liquidityFee", "platformFee"]
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO bet_fees ({fields}) VALUES ({placeholders})",
                fields=fee_fields,
                data=[{**bet["fees"], "betId": bet["id"]} for bet in bets if "fees" in bet]
            )

            # Delete entries
            prepare_and_execute_multi_deletion(
                conn=conn,
                query="DELETE FROM bet_fills WHERE betId = ?",
                ids=[bet["id"] for bet in bets]
            )

            fill_fields = ["betId", "timestamp", "matchedBetId", "amount", "shares"]
            prepare_and_execute_multi_upsert(
                conn=conn,
                query="INSERT OR REPLACE INTO bet_fills ({fields}) VALUES ({placeholders})",
                fields=fill_fields,
                data=[{**fill, "betId": bet["id"]} for bet in bets for fill in bet.get("fills", [])]

            )
            
            # Commit transaction
            conn.commit()

        except sqlite3.Error as e:
            logger.error("Database error in upsert_bets:", e)
            conn.rollback()  # Rollback transaction in case of error

        logger.debug("Upsert bets successful")

class ManifoldDatabaseReader:
    def __init__(self, manifold_db):
        self.manifold_db = manifold_db
        self.manifold_db.get_conn().row_factory = self.dict_factory

    def dict_factory(self, cursor, row):
        """Row factory to produce dictionary results."""
        return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}

    def execute_query(self, query, params=None):
        """
        Execute a read query and return the results.

        :param query: The SQL query string.
        :param params: Any parameters for the query (optional).
        :return: The query results.
        """
        conn = self.manifold_db.get_conn()
        
        logger.debug(f"Executing query {query} with params {params}")
        
        # Set row factory here so it returns dicts.
        conn.row_factory = self.dict_factory
        
        cursor = conn.cursor()
        cursor.execute(query, params or [])
        return cursor.fetchall()

class ManifoldDatabaseWriter:
    def __init__(self, manifold_db):
        self.manifold_db = manifold_db
        self.write_queue = queue.Queue()
        self.shutdown_flag = threading.Event()
        self.chunk_count = {}  # To track the number of chunks related to each future
        self.worker_thread = threading.Thread(target=self._write_thread, name="MF_DB_WRITE")
        self.worker_thread.start()

    def _write_thread(self):
        while not self.shutdown_flag.is_set():
            try:
                function, future, data = self.write_queue.get(timeout=1)
                try:
                    function(data)
                    
                    with threading.Lock():  # Ensure thread-safety when modifying chunk_count
                        self.chunk_count[future] -= 1
                        if self.chunk_count[future] == 0:
                            future.set_result(True)  # Indicate successful processing
                            del self.chunk_count[future]
                except Exception as e:
                    future.set_exception(e)
                    with threading.Lock():
                        del self.chunk_count[future]  # If an exception occurs, don't continue to count chunks
            except queue.Empty:
                continue

    def shutdown(self):
        logger.debug("Shutting down manifold database writer")
        self.shutdown_flag.set()
        self.worker_thread.join()

    def queue_write_operation(self, function, data):
        """
        Queue a write operation to the database.

        :param function: The function to execute (a write function from ManifoldDatabase class).
        :param data: The data to write.
        :return: Future object representing the execution of the operations.
        """
        logger.debug(f"Queueing write operation {function.__name__} with {len(data)} data items")
        
        sublists = self._split_list(data, 10000)
        future = concurrent.futures.Future()

        with threading.Lock():  # Ensure thread-safety when modifying chunk_count
            self.chunk_count[future] = len(sublists)

        for sublist in sublists:
            self.write_queue.put((function, future, sublist))
        
        return future

    def _split_list(self, input_list, max_size):
        """Splits the list into sublists of size <= max_size."""
        return [input_list[i:i + max_size] for i in range(0, len(input_list), max_size)]


from loguru import logger
from collections import defaultdict
from typing import List, Callable, Dict, DefaultDict
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.events import EVENT_JOB_EXECUTED, EVENT_JOB_ERROR
from autofold.api import ManifoldAPI
from autofold.database import ManifoldDatabase
from autofold.database import ManifoldDatabaseWriter



'''
TODO:

This whole thing has a litany of edge cases and problems when there are multiple strategies involved.
As it stands now it *works*, just not as robust as I would like.

Thoughts:
All callbacks associated with job should be fired. 
Right now it's possible for a subscriber to miss a callback if

Replace APScheduler with custom scheduler
- Job objects should be stored with function signature for proper sharing
- Job run interval should be min(polling_times) (coalesce jobs)
- Polling times per subscriber should be respected
'''

class ManifoldSubscriber():
	def __init__(self, manifold_api: ManifoldAPI, manifold_db: ManifoldDatabase, manifold_db_writer: ManifoldDatabaseWriter):
		self.manifold_api = manifold_api
		self.manifold_db = manifold_db
		self.manifold_db_writer = manifold_db_writer
		self.scheduler = BackgroundScheduler({

			'apscheduler.executors.default': {
				'class': 'apscheduler.executors.pool:ThreadPoolExecutor',
				'max_workers': '20'
			},
			'apscheduler.job_defaults.coalesce': 'true',
			'apscheduler.job_defaults.max_instances': '1',
		})
  
		# Dictionary to store job callbacks, defaulting to an empty list for each job ID
		self.callbacks: DefaultDict[str, List[Callable]] = defaultdict(list)
		
		# Add listener to the scheduler
		self.scheduler.add_listener(self._job_listener, EVENT_JOB_EXECUTED | EVENT_JOB_ERROR)
  
		self.scheduler.start()
  
	def shutdown(self):
		logger.debug("Shutting down manifold subscriber")
		self.scheduler.remove_listener(self._job_listener)
		self.scheduler.shutdown(wait=True)
		logger.debug("Manifold subscriber shut down")

	def _job_listener(self, event):
		'''
		Listens to job events to trigger registered callbacks
		'''
		job_id = event.job_id
		for callback in self.callbacks[job_id]:
			callback()

	def _get_job_key(self, func, args, polling_time):
		"""
		Generate a unique key for a job based on function, arguments, and polling time.
		"""
		return f"{func.__name__}_{str(args)}_{polling_time}"

	def _add_or_update_job(self, func, args, polling_time, callback):
		"""
		Check if a job with the given parameters already exists. If so, register the new callback.
		If not, create a new job.
		"""
		job_key = self._get_job_key(func, args, polling_time)
		if job_key in self.callbacks:
			self.register_callback(job_key, callback)
		else:
			job = self.scheduler.add_job(func=func, args=args, trigger='interval', seconds=polling_time)
			self.callbacks[job_key] = []
			self.register_callback(job_key, callback)

	def register_callback(self, job_id, callback):
		'''
		Register a callback function for a specific job ID
		'''
		logger.debug(f"Adding callback {callback.__name__} to job id {job_id}")
		self.callbacks[job_id].append(callback)

	def unregister_callback(self, job_id, callback):
		'''
		Unregister a specific callback function for a specific job ID
		'''
		if callback in self.callbacks[job_id]:
			logger.debug(f"Removing callback {callback.__name__} from job id {job_id}")
			self.callbacks[job_id].remove(callback)
			# If no more callbacks for this job ID, delete the job ID entry
			if not self.callbacks[job_id]:
				del self.callbacks[job_id]

	def subscribe_to_user_info(self, userId, polling_time=60, callback=None):
		'''
		Continuously retrieves the (LiteUser) profile of a user by their userid and updates the manifold database with it.
		
		:param polling_time: The number of seconds between updates. Default is 60 seconds.
		:param callback: Optional. The function to be called when the job finishes.
		:returns: job. The apscheduler job object.
		'''
		logger.debug(f"Subscribing to profile of user {userId} with polling time {polling_time} seconds")
		job = self.scheduler.add_job(func=self.update_user_info, args=(userId), trigger='interval', seconds=polling_time)
		if callback:
			self.register_callback(job.id, callback)
		return job
	
	def update_user_info(self, userId):
		'''
		Retrieves the (LiteUser) profile of a user by their userId and updates the manifold database with the fetched data.

		:param userId: The Id of the user to be fetched.
  
		NOTE: This function is blocking.
		'''
		logger.debug(f"Updating profile for user {userId}")
		user = self.manifold_api.get_user_by_id(userId=userId).result()
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_users, data=[user]).result()

	
	def subscribe_to_bets(self, userId=None, username=None, contractId=None, contractSlug=None, polling_time=60, callback=None):
		'''Continuously retrieves the bets via a single value or combination of:

		- userId
		- username
	 	- contractId
	 	- contractSlug 
	  
		and updates the manifold database with it.

		:param userId: The Id of the user.
		:param username: The username of the user.
		:param contractId: The Id of the contract (market).
		:param contractSlug: The URL slug of the contract (market). 
		:param polling_time: The number of seconds between updates. Default is 60 seconds.
		:param callback: Optional. The function to be called when the job finishes.
		:returns: job. The apscheduler job object.
		'''
		logger.debug(f"Subscribing to bets with userId={userId}, username={username}, contractId={contractId}, contractSlug={contractSlug}, and polling_time={polling_time} seconds with callback {callback.__name__ if callback else None}.")
		job = self.scheduler.add_job(func=self.update_user_bets, args=(userId, username, contractId, contractSlug), trigger='interval', seconds=polling_time)
		if callback:
			self.register_callback(job.id, callback)
		return job

	def update_bets(self, userId=None, username=None, contractId=None, contractSlug=None):
		'''
		Retrieves the bets of a user by their userId and updates the manifold database with the fetched data.

		:param userId: Optional. If set, the response will include only bets created by this user.
		:param username: Optional. If set, the response will include only bets created by this user.
		:param contractId: Optional. If set, the response will only include bets on this contract.
		:param contractSlug: Optional. If set, the response will only include bets on this contract.
  
		NOTE: This function is blocking.
		'''
		logger.debug(f"Updating bets with userId={userId}, username={username}, contractId={contractId} and contractSlug={contractSlug}")
  
		# Use locals to get the current local variables as a dictionary
		api_params = locals().copy()
		# Remove the 'self' key
		api_params.pop('self') 

		bets = self.manifold_api.retrieve_all_data(api_call_func=self.manifold_api.get_bets, max_limit=1000, **api_params)
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_bets, data=bets).result()

	
	def subscribe_to_market_positions(self, marketId, userId=None, polling_time=60, callback=None):
		''' 
		NOTE: Due to https://github.com/manifoldmarkets/manifold/issues/2031 this function will only subscribe to the top 4000 positions ranked by profit.

  		Continuously retrieves the positions of a market by its marketId and updates the manifold database with it. Optionally tracks a single user's positions.
	
		:param marketId: The id of the market
		:param userId: Optional. Tracks the positions of a specific user in a market.
		:param polling_time: The number of seconds between updates. Default is 60 seconds.
		:param callback: Optional. The function to be called when the job finishes.
		:returns: job. The apscheduler job object.
		'''
  
		logger.debug(f"Subscribing to market positions for marketId={marketId} and userId={userId} with a polling time of {polling_time} seconds and callback {callback.__name__ if callback else None}")
  
		job = self.scheduler.add_job(func=self.update_market_positions, args=(marketId, userId), trigger='interval', seconds=polling_time)
		if callback:
			self.register_callback(job.id, callback)
		return job

	def update_market_positions(self, marketId, userId=None):
		'''
		NOTE: Due to https://github.com/manifoldmarkets/manifold/issues/2031 this function will only retrieve to the top 4000 positions ranked by profit.

		Retrieves the positions of a market by its marketId, and optionally for a specific user, then updates the manifold database with the fetched data.

		:param marketId: The ID of the market whose positions are to be fetched.
		:param userId: Optional. If provided, fetches positions only for this user in the specified market.
  
		NOTE: This function is blocking.
		'''
  
		logger.debug(f"Updating market positios for marketId={marketId} and userId={userId}")
  
		contract_metrics = self.manifold_api.get_market_positions(marketId=marketId, order='profit', top=2000, userId=userId).result()
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_contract_metrics, data=contract_metrics).result()

	def subscribe_to_all_users(self, polling_time=3600, callback=None):
		''' 
		Continuously retrieves the (LiteUser) profile of all users and updates the manifold database with it.
		
		:param polling_time: The number of seconds between updates. Default is 3600 seconds.
		:param callback: Optional. The function to be called when the job finishes.
		:returns: job. The apscheduler job object.
		'''
	
		logger.debug(f"Subscribing to profiles of all users with polling time {polling_time} seconds and callback {callback.__name__ if callback else None}")
  
		job = self.scheduler.add_job(func=self.update_all_users, trigger='interval', seconds=polling_time)
		if callback:
			self.register_callback(job.id, callback)
		return job

	def update_all_users(self):
		'''
		Retrieves the (LiteUser) profile of all users and updates the manifold database with the fetched data.
		NOTE: This function is blocking.
		''' 
	
		logger.debug(f"Updating profiles of all users")
  
		users = self.manifold_api.retrieve_all_data(self.manifold_api.get_users, max_limit=1000)
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_users, data=users).result()

	def subscribe_to_all_markets(self, polling_time=3600, callback=None):
		''' 
  		Continuously retrieves all (LiteMarket) markets and updates the manifold database with it.
		
		:param polling_time: The number of seconds between updates. Default is 3600 seconds.
		:param callback: Optional. The function to be called when the job finishes.
		:returns: job. The apscheduler job object.
  
		NOTE: Currently only supports binary choice and multiple choice markets
		'''
  
		logger.debug(f"Subscribing to all markets with polling time {polling_time} seconds and callback {callback.__name__ if callback else None}")
  
		job = self.scheduler.add_job(func=self.update_all_markets, trigger='interval', seconds=polling_time)
		if callback:
			self.register_callback(job.id, callback)
		return job

	def update_all_markets(self):
		'''
		Retrieves all (LiteMarket) markets and updates the manifold database with the fetched data.
  
		NOTE: Currently only supports binary choice and multiple choice markets 
		NOTE: This function is blocking.
		''' 
	
		logger.debug("Updating all markets")
  
		markets = self.manifold_api.retrieve_all_data(self.manifold_api.get_markets, max_limit=1000)
		binary_choice_markets = []
		multiple_choice_markets = []
		for market in markets:
			if market["outcomeType"] == "BINARY":
				binary_choice_markets.append(market)
			elif market["outcomeType"] == "MULTIPLE_CHOICE":
				multiple_choice_markets.append(market)
   
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_binary_choice_markets, data=binary_choice_markets).result()
		self.manifold_db_writer.queue_write_operation(function=self.manifold_db.upsert_multiple_choice_markets, data=multiple_choice_markets).result()
  


from concurrent.futures import ThreadPoolExecutor
from loguru import logger
import signal
import threading
import time
from autofold.strategy import Strategy
from autofold.api import ManifoldAPI
from autofold.database import ManifoldDatabase
from autofold.database import ManifoldDatabaseWriter
from autofold.database import ManifoldDatabaseReader
from autofold.subscriber import ManifoldSubscriber



class StrategyBot:
	'''
	The StrategyBot is responsible for maintaining, adding, removing, starting and stopping strategies. 
 
	Attributes:
	-----------
	- ``manifold_api``: The ManifoldAPI instance
	- ``manifold_db``: The ManifoldDatabase instance
	- ``manifold_db_reader``: The ManifoldDatabaseReader instance
	- ``manifold_db_writer``: The ManifoldDatabaseWriter instance
	- ``manifold_subscriber``: The ManifoldSubscriber instance
	''' 
	def __init__(self):
	 
		# Bind Ctrl+C to the stop function
		signal.signal(signal.SIGINT, self.stop) 
	 
		self._strategies = []
  
		self.manifold_api = ManifoldAPI()

		self.manifold_db = ManifoldDatabase()
		self.manifold_db_reader = ManifoldDatabaseReader(self.manifold_db)
		self.manifold_db_writer = ManifoldDatabaseWriter(self.manifold_db)
		self.manifold_db.create_tables()
  
		self.manifold_subscriber = ManifoldSubscriber(self.manifold_api, self.manifold_db, self.manifold_db_writer)
 
		self.shutdown_event = threading.Event()  # Create an event object
  
 
	def register_strategy(self, strategy):
		'''
		Registers a trading strategy with the bot.

		:param type strategy: Required. The strategy class to register. Must be a subclass of the `Strategy` class.
		:raises TypeError: if the strategy is not a class type.
		:raises ValueError: if the strategy is not a subclass of `Strategy`.

		Example:

		.. code-block:: python

			bot = ManifoldBot()
			bot.register_strategy(MyCustomStrategy)
		''' 
		if not isinstance(strategy, Strategy):
			logger.error(f"{strategy} must be of a subclass of type Strategy")
			return

		self._strategies.append(strategy)
 	
	def start(self):
		'''
		Starts the bot's operation.

		Initializes the manifold API, databases, subscriber, and thread executor. Schedules the strategies to run in the thread pool.

		:raises RuntimeError: if no strategies have been registered.
		''' 
		self._executor = ThreadPoolExecutor(thread_name_prefix="BOT", max_workers=20) 
	
		if len(self._strategies) == 0:
			logger.error("No strategies have been registered for the bot. Exiting.")
			self.stop()
		
		# Schedule strategies to run in the thread pool
		self.futures = [self._executor.submit(strategy.start) for strategy in self._strategies]
  
		# Loop to check for the stop condition
		while not self.shutdown_event.is_set():
			time.sleep(1)  # Sleep for 1 second

	def stop(self, *args):
		'''
		Stops the bot's operation.

		Shuts down the API, stops all running strategies, and shuts down the thread pool and subscriber.
		''' 
 		# API must be shut down first
		self.manifold_api.shutdown()

		for strategy in self._strategies:
			strategy.stop()
		self._executor.shutdown(wait=False)
  
		# Subscriber
		self.manifold_subscriber.shutdown()

		# Database 
		self.manifold_db_writer.shutdown()  
  
		self.shutdown_event.set() 


	def get_id(self):
		# Set bot id
		self.me = self.manifold_api.get_me().result()
		self.id = self.me["id"]



# from fractions import Fraction
# from functools import reduce
# from operator import add
# import math

# CREATOR_FEE = ...  # Add appropriate value
# LIQUIDITY_FEE = ...  # Add appropriate value
# PLATFORM_FEE = ...  # Add appropriate value
# EPSILON = 0.00000001  # Add appropriate value

# class LimitBet:
#     ...

# class LiquidityProvision:
#     ...

# class Answer:
#     ...


# def get_cpmm_probability(pool, p):
#     YES, NO = pool['YES'], pool['NO']
#     return (p * NO) / ((1 - p) * YES + p * NO)


# def get_cpmm_probability_after_bet_before_fees(state, outcome, bet):
#     pool, p = state['pool'], state['p']
#     shares = calculate_cpmm_shares(pool, p, bet, outcome)
#     y, n = pool['YES'], pool['NO']

#     if outcome == 'YES':
#         newY, newN = y - shares + bet, n + bet
#     else:
#         newY, newN = y + bet, n - shares + bet

#     return get_cpmm_probability({'YES': newY, 'NO': newN}, p)


# def get_cpmm_outcome_probability_after_bet(state, outcome, bet):
#     new_pool = calculate_cpmm_purchase(state, bet, outcome)['newPool']
#     p = get_cpmm_probability(new_pool, state['p'])
#     return 1 - p if outcome == 'NO' else p


# def calculate_cpmm_shares(pool, p, bet, bet_choice):
#     if bet == 0:
#         return 0

#     y, n = pool['YES'], pool['NO']
#     k = y ** p * n ** (1 - p)

#     if bet_choice == 'YES':
#         return y + bet - (k * (bet + n) ** (p - 1)) ** (1 / p)
#     else:
#         return n + bet - (k * (bet + y) ** -p) ** (1 / (1 - p))


# def get_cpmm_fees(state, bet, outcome):
#     prob = get_cpmm_probability_after_bet_before_fees(state, outcome, bet)
#     bet_p = 1 - prob if outcome == 'YES' else prob

#     liquidity_fee = LIQUIDITY_FEE * bet_p * bet
#     platform_fee = PLATFORM_FEE * bet_p * bet
#     creator_fee = CREATOR_FEE * bet_p * bet
#     fees = {'liquidityFee': liquidity_fee, 'platformFee': platform_fee, 'creatorFee': creator_fee}

#     total_fees = liquidity_fee + platform_fee + creator_fee
#     remaining_bet = bet - total_fees

#     return {'remainingBet': remaining_bet, 'totalFees': total_fees, 'fees': fees}

# # ... [Continue for the rest of the functions]

# # Note: Python lacks direct translations for some JavaScript functions such as `mapValues` and `sumBy` from lodash.
# #       You might need to implement those utility functions yourself or use Pythonic methods to achieve the same result.



from loguru import logger
from tinydb import where
from autofold.strategy import Strategy
from autofold.bot import StrategyBot


class ExampleStrategy(Strategy):
	def __init__(self, bot: StrategyBot):
		super().__init__(__name__, bot)
		logger.info("ExampleStrategy object initialized successfully.") 
		
	
	def start(self):
		self.running = True
		logger.info("Running the strategy")
		'''
			In this example:
				- Find the top 10 unresolved binary choice markets with the highest daily volume
				- Find the user with the highest profit percent position of the 10 markets
				- Find their most recent bet and the market it was in
				- Make a bet in the direction of their last bet in the market
				- Subscribe to the user position in the market and do what they do
				Includes:
				- Persistence between strategy executions

			
			NOTE: This is just an example to demonstrate how to use this system. There are a lot of problems/edge cases with this. 
   				  A real strategy might be much more sophisticated. You will probably lose mana with this strategy!!!
		'''
		init_status = self.db.search(where('init').exists())
		if not init_status:
			logger.info("Strategy is not initialized. Initializing now.") 
			self.init_strategy()
		else:
			logger.info("Strategy has already been initialized.") 
		   
			# Get the ids for our stuff from the db 
			self.best_position_db_id = self.db.get(where('best_position').exists()).doc_id
			self.best_position_shares_db_id = self.db.get(where('best_position_shares').exists()).doc_id
			
			best_position = self.db.get(doc_id=self.best_position_db_id)["best_position"]
			
			logger.info(f"Subscribing to future bets of user {best_position['userName']} in market {best_position['contractId']} with polling time {60} seconds")
			self.manifold_subscriber.subscribe_to_market_positions(marketId=best_position["contractId"],
															   userId=best_position["userId"],
															   polling_time=60,
															   callback=self.track_position) 
	
	def stop(self):
		self.running = False
		logger.info("Shutdown method called. Strategy has been halted.") 
		
	def init_strategy(self):
		logger.info("Initializing the strategy.") 

		# clear db
		self.db.truncate() 
		logger.debug("Database truncated.") 

		# Update the manifold database with all markets
		logger.info("Updating the manifold database with all markets. (This may take a while)") 
		# self.manifold_subscriber.update_all_markets()

		# Find top 10 binary choice markets with highest volume 
		logger.debug("Finding top 10 unresolved markets with highest 24hr volume")
		markets = \
			self.manifold_db_reader.execute_query(
				"""
				SELECT 
					id,
					volume24Hours,
					question,
					url
				FROM 
					binary_choice_markets
				WHERE
					isResolved = FALSE
				ORDER BY 
					volume24Hours DESC
				LIMIT 10;
				""")
		if not markets:
			logger.error("Failed to fetch top 10 binary choice markets.")
		else:
			logger.success(f"Found top 10 markets by 24 hr volume") 
			logger.debug(f"Markets: {markets}")

		# Save them to local db
		self.markets_db_id = self.db.insert({"markets": markets}) 
		logger.debug(f"Saved markets to db with id: {self.markets_db_id}") 
		
		# Update the manifold database with all the positions (contractMetrics) associated with these markets
		logger.info("Fetching positions for each of the top 10 markets")
		for market in markets:
			logger.debug(f"Fetching positions for market {market['id']}")
			self.manifold_subscriber.update_market_positions(marketId=market["id"])
		   
		# Extract market ids
		market_ids = [market["id"] for market in markets]
		
		# Find the highest profit percent position
		logger.debug("Finding best position (highest profit percent) from top 10 markets.")
		best_position = \
			self.manifold_db_reader.execute_query(
				"""
				SELECT 
					contractId,
					userId,
					userUsername,
					userName,
					profitPercent
				FROM 
					contract_metrics
				WHERE 
					contractId IN (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
				ORDER BY 
					profitPercent DESC
				LIMIT 1;
				""", market_ids)[0]
	  
		if not best_position:
			logger.error("Failed to fetch best position.")
		else:
			logger.success(f"Found best position") 
			logger.debug(f"Best position: {best_position}")

		# Save it to the db
		self.best_position_db_id = self.db.insert({"best_position": best_position})
		logger.debug(f"Saved best position to db with id: {self.best_position_db_id}") 
 
		
		# Find the YES/NO shares for this position
		logger.debug("Finding the YES/NO shares for the best position")
		best_position_shares = \
			self.manifold_db_reader.execute_query(
				"""
				SELECT 
					outcome,
					numberOfShares
				FROM 
					contract_metrics_totalShares
				WHERE 
					userId = ? AND contractId = ?
				ORDER BY 
					id DESC
				LIMIT 2;  -- Since there are only two outcomes, YES and NO
				""", (best_position["userId"], best_position["contractId"]))

		if not best_position_shares:
			logger.error("Failed to find YES/NO shares for best position.")
		else:
			logger.info(f"Found YES/NO shares for best position: {best_position_shares}") 
	
	 
		# Save it to the db
		self.best_position_shares_db_id = self.db.insert({"best_position_shares": best_position_shares})
		logger.debug(f"Saved best position YES/NO shares to db with id: {self.best_position_shares_db_id}")
	   
		# Update the bets from the user with the best position
		logger.info(f"Fetching bets for user {best_position['userName']}")
		self.manifold_subscriber.update_bets(userId=best_position["userId"], contractId=best_position["contractId"])
		
		# Find the most recent bet from the user with the best position
		logger.debug(f"Finding most recent bet for user {best_position['userId']}")
		recent_bet = \
			self.manifold_db_reader.execute_query(
				"""
				SELECT 
					id,
					userId,
					contractId,
					amount,
					outcome,
					createdTime
				FROM 
					bets
				WHERE 
					userId = ? AND contractId = ?
				ORDER BY 
					createdTime DESC
				LIMIT 1;
				""", (best_position["userId"], best_position["contractId"]))[0] 

		if not recent_bet:
			logger.error(f"Failed to fetch most recent bet from user {best_position['userId']} ")
		else:
			logger.info(f"Found most recent bet for user {best_position['userName']}")

		# Save the recent bet to the db
		self.recent_bet_db_id = self.db.insert({"recent_bet": recent_bet})
		logger.debug(f"Saved most recent bet to db with id: {self.recent_bet_db_id}")
		
			
		
		# Make a bet in this direction
		# NOTE: From the API docs:
		# "A non-refundable transaction fee of M0.25 will be levied on any bet, sell, or limit order placed through the API, or by any account marked as a bot."
		# NOTE: The manifold_api returns future objects for each request. You can block and wait for them to finish by calling .result() on them.
		print(f"self.manifold_api.make_bet(amount=10, contractId={recent_bet['contractId']}, outcome={recent_bet['outcome']}).result()")
		
		# NOTE: Manifold uses 'contractId' and 'marketId' interchangably. 
		# 'marketId' is used when referring directly to a market
		# 'contractId' is used _within_ bets or positions to refer to the market it is in
		# :(
		
		# Subscribe to any future changes in position
		logger.info(f"Subscribing to future bets of user {best_position['userName']} in market {recent_bet['contractId']} with polling time {60} seconds")
		self.manifold_subscriber.subscribe_to_market_positions(marketId=recent_bet["contractId"],
															   userId=recent_bet["userId"],
															   polling_time=60,
															   callback=self.track_position)
		
		# Done!
		self.db.insert({'init': True}) 

	# This will execute every 60 seconds(or polling_time) after the position from the user in the specified market have been updated into the manifold db
	def track_position(self):
		# Retrieve the best position and recent bet data from the local db
		best_position = self.db.get(doc_id=self.best_position_db_id)["best_position"]
		best_position_shares = self.db.get(doc_id=self.best_position_shares_db_id)["best_position_shares"]

		# Fetch the latest shares (YES and NO) from the nested contract_metrics_totalShares
		latest_shares = \
			self.manifold_db_reader.execute_query(
				"""
				SELECT 
					outcome,
					numberOfShares
				FROM 
					contract_metrics_totalShares
				WHERE 
					userId = ? AND contractId = ?
				ORDER BY 
					id DESC
				LIMIT 2;  -- Since there are only two outcomes, YES and NO
				""", (best_position["userId"], best_position["contractId"]))
		
		# Do what they do if their position changes from what we have
		# Convert shares list to dictionary
		latest_shares_dict = {item['outcome']: item['numberOfShares'] for item in latest_shares}
		best_position_shares_dict = {item['outcome']: item['numberOfShares'] for item in best_position_shares}

		# 1. Check for new positions
		for outcome, shares in latest_shares_dict.items():
			if outcome not in best_position_shares_dict:
				logger.info(f"New position taken: {outcome} with {shares} shares.")
				# Place a bet for the new position
				# self.manifold_api.make_bet(amount=shares, contractId=best_position["contractId"], outcome=outcome)  # you can add limitProb, expiresAt, etc. if needed

		# 2. Check for sold positions
		for outcome, shares in best_position_shares_dict.items():
			if outcome not in latest_shares_dict:
				logger.info(f"Entire position sold: {outcome} with {shares} shares.")
				# Sell the shares for the sold position
				# self.manifold_api.sell_shares(marketId=best_position["contractId"], outcome=outcome, shares=shares)

		# 3. Check for buy/sell actions on existing positions
		for outcome, shares in latest_shares_dict.items():
			if outcome in best_position_shares_dict:
				difference = shares - best_position_shares_dict[outcome]
				if difference > 0:
					logger.info(f"Bought {difference} shares of {outcome}.")
					# Place a bet for the bought shares
					self.manifold_api.make_bet(amount=difference, contractId=best_position["contractId"], outcome=outcome)
				elif difference < 0:
					logger.info(f"Sold {-difference} shares of {outcome}.")
					# Sell the sold shares
					# self.manifold_api.sell_shares(marketId=best_position["contractId"], outcome=outcome, shares=-difference)

def main():

	# Set up logging
	log_format = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"

	# Configure the logger
	logger.add("logs/{time:YYYY-MM-DD}.log", rotation="1 day", format=log_format, level="TRACE", enqueue=True) 
	logger.info("Logging has been set up!") 

	# Init bot
	strategy_bot = StrategyBot()

	# Init strategy
	example_strategy = ExampleStrategy(strategy_bot)
	
	# Register it
	strategy_bot.register_strategy(example_strategy)

	# Run the bot
	strategy_bot.start()

main() 
  